#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Contains the Jinja templates to generate javascript.
"""

__version__ = "1.0.0"
"""Version of the js template set."""
