# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# This software is distributed under the terms of the MIT License.

"""
Contains the Jinja templates to generate Py headers.
"""

__version__ = "1.0.0"
"""Version of the py template set."""
